/* wblock.c -- Copyright 1991, 1993 Liam R. E. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/* The low-level physical Word Database for lq-text.
 *
 * $Id: wblock.c,v 1.4 93/11/28 15:17:25 lee Exp $
 */

#include "error.h"

#include <stdio.h> /* stderr, also for fileinfo.h */
#include <fcntl.h>
#include <sys/types.h>

#include "globals.h" /* defines and declarations for database filenames */

#include "fileinfo.h" /* for wordinfo.h */
#include "wordinfo.h"
#include "pblock.h"
#include "numbers.h"
#include "wordrules.h"
#include "putbyte.h"

extern t_WordInfo *WID2WordInfo();
extern t_WID Word2WID();

/** C library functions that need to be declared: **/

/** lqtext library functions that need to be declared: **/
extern unsigned int FindFreeBlock();

/** Functions within this file that need to be declared: **/

/** **/

#ifdef ASCIITRACE
extern int AsciiTrace;
#endif

extern LIBRARY char *ReadBlock();
extern LIBRARY void WriteBlock();

#include "blkheader.h"

static unsigned char pblockBuffer[8192 * 2]; /*TODO use CACHELEN */
/* TODO: malloc() pblockBuffer or (better) eliminate it altogether */

API void
FlushBlock(Block, ByteCount, NextOffset, LastStart, WID)
    char *Block;
    int ByteCount;
    unsigned long *NextOffset, *LastStart;
    t_WID WID;
{
    if (*LastStart && Block) {
	/*NOSTRICT*/
	t_BlockHeader *BH = (t_BlockHeader *) Block;
	register char *p;
	int BlocksToWrite = (ByteCount + BLOCKSIZE - 1) / BLOCKSIZE;
							/* round up! */

	BH->NextOffset = (*NextOffset);
#ifdef WIDINBLOCK
	BH->WID = WID;
#endif
	/* pad with -1 for future compatibility */
	for (p = &Block[BLOCKSIZE*BlocksToWrite - 1]; p >= &Block[ByteCount]; p--) {
	     *(unsigned char *)p = (unsigned char) 0xFF;
	}
	WriteBlock(*LastStart, Block, BlocksToWrite);
    }

    if (*NextOffset) {
	char *p;
	t_BlockHeader *BH;

	Error(E_INTERNAL|E_WARN, "FlushBlock %ld != 0 (WID %ld)",
					*NextOffset, WID);
	p = ReadBlock(*NextOffset);
	BH = (t_BlockHeader *) Block;
	BH->NextOffset = 0L;
#ifdef WIDINBLOCK
	BH->WID = WID;
#endif
	/* write just the one block. */
	WriteBlock(*NextOffset, Block, 1);
    }
    *LastStart = *NextOffset = 0L;
}

/* This is simply to help keep the source lines getting too long! */
typedef unsigned char *UCP;

LIBRARY int
_PutByte(Byte, WID, sp, Blockp, BlockLength, LastStart, NextBlock, NextLength)
    unsigned char Byte;
    t_WID WID;
    unsigned char **sp;
    unsigned char **Blockp;
    unsigned int* BlockLength;
    unsigned long *NextBlock;
    unsigned long *NextLength;
    unsigned long *LastStart; /* for writing the linked list */
{
    t_BlockHeader *BH;

    if (*sp - (*Blockp) >= (*BlockLength)) {
	int NumberOfBlocks = (*BlockLength + BLOCKSIZE - 1) / BLOCKSIZE;

	if (!*NextBlock && !*LastStart) return -1; /* only do the 1st block */
	if (*NextBlock == (unsigned long) 0) {
	    *NextBlock = FindFreeBlock(WID, BlockLength);
	} else {
	    if (NextLength && *NextLength) {
		*BlockLength = (*NextLength);
		*NextLength = 0;
	    } else {
		Error(E_BUG,
		    "_PutByte: NextLength is %s",
		    NextLength ? "zero" : "a null pointer"
		);
	    }
	}

	/* Complete the information in the previous block, if required */
	if (*LastStart) {
	    BH = (t_BlockHeader *) (*Blockp);
#ifdef WIDINBLOCK
	    BH->WID = WID;
#endif
	    BH->NextOffset = (*NextBlock);
	    /* Write the old block */
	    WriteBlock(*LastStart, *Blockp, NumberOfBlocks);
	    *LastStart = 0L;
	}
	*LastStart = (*NextBlock);
	(*NextBlock) = 0L; /* after this point NextLength is ignored */
	*Blockp = pblockBuffer; /* Use static (to this file) data buffer */
	/*NOSTRICT*/
	BH = (t_BlockHeader *) (*Blockp);
	BH->NumberOfBlocks = (unsigned char) (*BlockLength/BLOCKSIZE); /*exact*/
	(*sp) = (UCP) BH->Data;
    }
    **sp = Byte;
    (*sp)++;
    return 0;
}

/* PutLong -- write a long number in compressed/abbreviated form into a
 * string.  If this moves the string pointer beyond the block, write out
 * the block and start a new one.  In that case, the number written may well
 * span the gap between the blocks.  We use an overflow buffer to copy
 * the bytes (if any) that overflowed into it.
 * Then we write them at the start of the next block.
 *
 * This routine returns -1 and writes a partial number (no allocated block)
 * if *LastBlock and *NextBlock are zero.  This allows PutwOrdPlaces to be
 * called to put the WordPlaces into the WIDFILE block without writing out
 * an entire chain.
 */

LIBRARY int
_PutLong(Long, WID, sp, Blockp, BlockLength, LastStart, NextBlock, NextLength)
    unsigned long Long;
    t_WID WID;
    unsigned char **sp;
    unsigned char **Blockp;
    unsigned *BlockLength;
    unsigned long *NextBlock;
    unsigned long *NextLength;
    unsigned long *LastStart; /* for writing the linked list */
{
    unsigned char *Start = (*sp);
 
    if (sWriteNumber(
	(char **) sp,
	Long, 
	*Blockp,
	*BlockLength
    ) < 0) {
	unsigned int BytesWritten;
        int NumberOfBlocks;
	t_BlockHeader *BH;
 
        if (!*NextBlock && !*LastStart) {
	    /* We were being tentative, seeing if the data would fit
	     * or not, and it didn't.
	     */
	    return -1;
	}
 
        NumberOfBlocks = (*BlockLength + BLOCKSIZE - 1) / BLOCKSIZE;
	BytesWritten = (*BlockLength - (Start - (*Blockp)));
 
        if (*NextBlock == (unsigned long) 0) {
            *NextBlock = FindFreeBlock(WID, BlockLength);
        } else {
            if (NextLength && *NextLength) {
                *BlockLength = (*NextLength);
                *NextLength = 0;
            } else {
                Error(E_BUG,
                    "_PutLong: NextLength is %s",
                    NextLength ? "zero" : "a null pointer"
                );
            }
        }
        /* Complete the information in the previous block, if required */
        if (*LastStart) {
            BH = (t_BlockHeader *) (*Blockp);
            BH->NextOffset = *NextBlock;
#ifdef WIDINBLOCK
            BH->WID = WID;
#endif
            /* Write the old block */
            WriteBlock(*LastStart, *Blockp, NumberOfBlocks);
        }
        *LastStart = (*NextBlock);
        (*NextBlock) = 0L;
        *Blockp = pblockBuffer;
        BH = (t_BlockHeader *) (*Blockp);
        (*sp) = (UCP) BH->Data;
        BH->NumberOfBlocks = *BlockLength / BLOCKSIZE; /* always exact */

	/* now write the rest of the number out */
	{
	    unsigned char Buffer[sizeof(unsigned long) + 2];
	    unsigned char *Bufp = Buffer;
	    register unsigned char *p;

	    (void) sWriteNumber((char **) &Bufp, Long, Buffer, sizeof Buffer);

            for (p = &Buffer[BytesWritten]; p < Bufp; p++) {
                *((*sp)++) = (*p);
            }
        }
    }
    return 0;
}
